@@ -33,14 +33,13 @@ import numpy as np
 from osgeo import gdal, gdal_array, gdalnumeric
 from shapely.geometry import Polygon
 from shapely.wkt import loads as shply_loads
-from six import PY3
 # dill -> imported when dumping GeoArray
 
 from py_tools_ds.convenience.object_oriented import alias_property
 from py_tools_ds.geo.coord_calc import get_corner_coordinates
 from py_tools_ds.geo.coord_grid import snap_bounds_to_pixGrid
 from py_tools_ds.geo.coord_trafo import mapXY2imXY, imXY2mapXY, transform_any_prj, reproject_shapelyGeometry
-from py_tools_ds.geo.projection import prj_equal, WKT2EPSG, EPSG2WKT, isLocal
+from py_tools_ds.geo.projection import prj_equal, WKT2EPSG, EPSG2WKT, isLocal, CRS
 from py_tools_ds.geo.raster.conversion import raster2polygon
 from py_tools_ds.geo.vector.topology \
     import get_footprint_polygon, polyVertices_outside_poly, fill_holes_within_poly
@@ -53,12 +52,6 @@ from py_tools_ds.numeric.array import get_array_tilebounds
 from .subsetting import get_array_at_mapPos
 from .metadata import GDAL_Metadata
 
-if PY3:
-    # noinspection PyCompatibility
-    from builtins import TimeoutError, FileNotFoundError
-else:
-    from py_tools_ds.compatibility.python.exceptions import TimeoutError, FileNotFoundError
-
 __author__ = 'Daniel Scheffler'
 
 
@@ -85,7 +78,6 @@ class GeoArray(object):
         :param progress:        show progress bars (default: True)
         :param q:               quiet mode (default: False)
         """
-
         # TODO implement compatibility to GDAL VRTs
         if not (isinstance(path_or_array, (str, np.ndarray, GeoArray)) or
            issubclass(getattr(path_or_array, '__class__'), GeoArray)):
@@ -142,6 +134,18 @@ class GeoArray(object):
             if self.filePath:
                 self.set_gdalDataset_meta()
 
+            if 'nodata' in self._initParams and self._initParams['nodata'] is not None:
+                self._validate_nodataVal()
+
+    def _validate_nodataVal(self):
+        """Check if a given nodata value is within the valid value range of the data type."""
+        if (np.issubdtype(self.dtype, np.integer) and
+           not np.can_cast(self._initParams['nodata'], self.dtype)) or \
+           (np.issubdtype(self.dtype, np.floating) and
+           not np.can_cast(float(self._initParams['nodata']), self.dtype)):
+            raise ValueError("The given no-data value (%s) is out range for data type %s."
+                             % (self._initParams['nodata'], self.dtype))
+
     @property
     def arr(self):
         return self._arr
@@ -183,7 +187,7 @@ class GeoArray(object):
             bN_dict = OrderedDict((band, i) for i, band in enumerate(list_bandnames))
 
             if len(bN_dict) != self.bands:
-                raise ValueError('Bands must different names. Received band list: %s' % list_bandnames)
+                raise ValueError('Bands must have unique names. Received band list: %s' % list_bandnames)
 
             self._bandnames = bN_dict
 
@@ -226,7 +230,6 @@ class GeoArray(object):
     @property
     def rows(self):
         """Get the number of rows of the associated image array."""
-
         return self.shape[0]
 
     @property
@@ -240,13 +243,11 @@ class GeoArray(object):
     @property
     def bands(self):
         """Get the number of bands of the associated image array."""
-
         return self.shape[2] if len(self.shape) > 2 else 1
 
     @property
     def dtype(self):
         """Get the numpy data type of the associated image array."""
-
         if self._dtype:
             return self._dtype
         elif self.is_inmem:
@@ -258,7 +259,6 @@ class GeoArray(object):
     @property
     def geotransform(self):
         """Get the GDAL GeoTransform of the associated image, e.g., (283500.0, 5.0, 0.0, 4464500.0, 0.0, -5.0)"""
-
         if self._geotransform:
             return self._geotransform
         elif not self.is_inmem:
@@ -283,13 +283,11 @@ class GeoArray(object):
     @property
     def xgsd(self):
         """Get the X resolution in units of the given or detected projection."""
-
         return self.geotransform[1]
 
     @property
     def ygsd(self):
         """Get the Y resolution in units of the given or detected projection."""
-
         return abs(self.geotransform[5])
 
     @property
@@ -326,8 +324,8 @@ class GeoArray(object):
 
     @property
     def epsg(self):
+        # type: (None) -> int
         """Get the EPSG code of the projection of the GeoArray."""
-
         return WKT2EPSG(self.projection)
 
     @epsg.setter
@@ -483,7 +481,7 @@ class GeoArray(object):
             else:
                 try:
                     multipolygon = raster2polygon(self.mask_nodata.astype(np.uint8), self.gt, self.prj, exact=False,
-                                                  progress=self.progress, q=self.q, maxfeatCount=10, timeout=3)
+                                                  progress=self.progress, q=self.q, maxfeatCount=10, timeout=5)
                     self._footprint_poly = fill_holes_within_poly(multipolygon)
                 except (RuntimeError, TimeoutError):
                     if not self.q:
@@ -791,10 +789,9 @@ class GeoArray(object):
             # for some reason GDAL reads arbitrary geotransforms as (0, 1, 0, 0, 0, 1) instead of (0, 1, 0, 0, 0, -1)
             self._geotransform[5] = -abs(self._geotransform[5])  # => force ygsd to be negative
 
-            # temp conversion to EPSG needed because GDAL seems to modify WKT string when writing file to disk
-            # (e.g. using gdal_merge) -> conversion to EPSG and back undos that
+            # consequently use WKT1 strings here as GDAL always exports transformation results as WKT1
             wkt = ds.GetProjection()
-            self._projection = EPSG2WKT(WKT2EPSG(wkt)) if not isLocal(wkt) else ''
+            self._projection = CRS(wkt).to_wkt(version="WKT1_GDAL") if not isLocal(wkt) else ''
 
             if 'nodata' not in self._initParams or self._initParams['nodata'] is None:
                 band = ds.GetRasterBand(1)
@@ -1343,10 +1340,14 @@ class GeoArray(object):
         assert self.projection, "A projection is needed for a map visualization. Got '%s'." % self.projection
 
         # get image to plot
+        # (reproject to LonLat as workaround in case self.epsg is None because cartopy relies on an existing EPSG code)
         nodataVal = nodataVal if nodataVal is not None else self.nodata
-        image2plot, gt, prj = self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
-                                                        boundsMapPrj=boundsMapPrj, res_factor=res_factor,
-                                                        nodataVal=nodataVal)
+        gA2plot = GeoArray(*self._get_plottable_image(xlim, ylim, band, boundsMap=boundsMap,
+                                                      boundsMapPrj=boundsMapPrj, res_factor=res_factor,
+                                                      nodataVal=nodataVal,
+                                                      out_prj=None if self.epsg is not None else 4326),
+                           nodata=nodataVal)
+        image2plot = gA2plot[:]
 
         # create map
         def get_cartopy_crs_from_epsg(epsg_code):
@@ -1358,19 +1359,21 @@ class GeoArray(object):
                 else:
                     raise NotImplementedError('The show_map() method currently does not support the given projection.')
 
-        crs_in = get_cartopy_crs_from_epsg(self.epsg)
-        crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else self.epsg)
+        crs_in = get_cartopy_crs_from_epsg(gA2plot.epsg)
+        crs_out = get_cartopy_crs_from_epsg(out_epsg if out_epsg is not None else gA2plot.epsg)
 
         fig = plt.figure(figsize=figsize)
         ax = plt.axes(projection=crs_out)
 
-        ax.set_extent(self.box.boundsMap, crs=crs_in)
+        ax.set_extent(gA2plot.box.boundsMap, crs=crs_in)
 
-        palette, vmin, vmax = self._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
+        palette, vmin, vmax = gA2plot._get_cmap_vmin_vmax(cmap, vmin, vmax, pmin, pmax, image2plot, nodataVal)
         if nodataVal is not None and np.std(image2plot) != 0:  # do not show nodata
             image2plot = np.ma.masked_equal(image2plot, nodataVal)
         ax.imshow(image2plot, cmap=palette, interpolation=interpolation, vmin=vmin, vmax=vmax,
-                  origin='upper', transform=crs_in, extent=list(self.box.boundsMap))
+                  origin='upper', transform=crs_in,
+                  extent=list(gA2plot.box.boundsMap)
+                  )
 
         # show grid lines
         try:
@@ -1403,7 +1406,7 @@ class GeoArray(object):
         import folium
         import geojson
 
-        lonlatPoly = reproject_shapelyGeometry(self.footprint_poly, self.epsg, 4326)
+        lonlatPoly = reproject_shapelyGeometry(self.footprint_poly, self.prj, 4326)
 
         m = folium.Map(location=tuple(np.array(lonlatPoly.centroid.coords.xy).flatten())[::-1])
         gjs = geojson.Feature(geometry=lonlatPoly, properties={})
